import subprocess
import threading

from flyers import assets
from flyers.logs import logger


class RunShellResult(object):
    def __init__(self, output, status, message, stdout=None, stderr=None):
        self.output = output
        self.status = status
        self.message = message
        self.stdout = stdout
        self.stderr = stderr


def run_shell(cmd: str) -> bool:
    assets.check_blank(cmd, 'cmd')
    return subprocess.call(cmd, shell=True) == 0


def batch_run_shell(cmds: list[str], break_on_fail: bool = False) -> list[bool]:
    result = []
    for cmd in cmds:
        try:
            ret = run_shell(cmd)
        except Exception:
            if break_on_fail:
                return result + [False]  # add the last one
            ret = False
        result.append(ret)
    return []


def async_run_shell(cmd: str, callback=None) -> str:
    """
    Executes command asynchronously
    :param cmd: Command
    :param callback:
    :return: Thread name
    """
    assets.check_blank(cmd, 'cmd')

    def runnable():
        ret = subprocess.call(cmd, shell=True) == 0
        if callback:
            callback(ret)

    t = threading.Thread(target=runnable)
    t.start()
    return t.getName()


def run_shell_with_result(cmd: str) -> (str, str):
    assets.check_blank(cmd, 'cmd')

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    output = process.communicate()
    if len(output) != 2:
        return RunShellResult(output, status=-1, message='The output is not expected')

    stdout, stderr = str(output[0], encoding='utf-8'), str(output[1], encoding='utf-8')

    logger.debug("Run <{}> command, stdout: {}".format(cmd, stdout))
    logger.debug("Run <{}> command, stderr: {}".format(cmd, stderr))
    return RunShellResult(output, status=0, message=None, stdout=stdout, stderr=stderr)


def async_run_shell_with_result(cmd, callback):
    def listener():
        result = run_shell(cmd)
        callback(result)

    t = threading.Thread(target=listener)
    t.start()



import pymysql
import pymysql.cursors

from flyers.database import DatabaseConf
from flyers.logs import logger


def execute(conn: pymysql.Connection, consumer_func, auto_commit=True):
    with conn.cursor() as cursor:
        consumer_func(cursor)
        if auto_commit:
            conn.commit()


def create_connection(conf: DatabaseConf) -> pymysql.Connection:
    try:
        return pymysql.connect(host=conf.host,
                               port=conf.port,
                               user=conf.user,
                               password=conf.password,
                               database=conf.database,
                               cursorclass=pymysql.cursors.DictCursor)
    except Exception:
        msg = 'Init connection error, conf: {}'.format(conf.__dict__)
        logger.exception(msg)
        raise Exception(msg)


def show_mysql_server_version(conn: pymysql.Connection):
    with conn.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        return cursor.fetchone()



import pymysql
from dbutils.pooled_db import PooledDB

from flyers.database import DatabaseConf
from flyers.logs import logger

# The max number of connections
MAX_CONNECTIONS = 10


def new_connection_pool(conf: DatabaseConf) -> PooledDB:
    """
    Create a new connection pool
    :param conf:
    :return:
    """
    logger.info('create new database connection pool, conf: {}'.format(conf))
    return PooledDB(
        pymysql,
        MAX_CONNECTIONS,
        host=conf.host,
        user=conf.user,
        port=conf.host,
        passwd=conf.password,
        db=conf.database,
        use_unicode=True)


def run_with(pool: PooledDB, consumer_func):
    conn, cursor = get_conn(pool)
    consumer_func(conn, cursor)
    close_conn(conn, cursor)


def get_conn(pool: PooledDB):
    conn = pool.connection()
    return conn, conn.cursor()


def close_conn(conn, cursor):
    conn.close()
    cursor.close()



import unittest

from flyers import bools


class MyTestCase(unittest.TestCase):
    def test_something(self):
        self.assertTrue(bools.is_true(1))
        self.assertTrue(bools.is_true('1'))
        self.assertTrue(bools.is_true('True'))
        self.assertTrue(bools.is_true('true'))

        self.assertFalse(bools.is_true(0))
        self.assertFalse(bools.is_true('false'))


if __name__ == '__main__':
    unittest.main()



def is_emtpy(s: str):
    return s is None or len(s) == 0


def is_blank(s: str):
    if not s:
        return True
    return is_emtpy(s.lstrip().rstrip())


def repeat(s: str, n: int):
    return s * n



from flyers.datetimes import *
import unittest


class MyTestCase(unittest.TestCase):
    def test_something(self):
        t = 1624170600
        d = datetime.fromtimestamp(t)

        self.assertEqual(timestamp_to_date_str(t), '2021-06-20')
        self.assertEqual(timestamp_to_datetime_str(t), '2021-06-20 14:30:00')

        self.assertEqual(parse_datetime('2021-06-20 14:30:00'), d)
        self.assertEqual(parse_date('2021-06-20'), datetime.fromtimestamp(1624118400))

        self.assertFalse(check_datetime('2021-002-1212'))

        print(format_duration(1657247354000))


if __name__ == '__main__':
    unittest.main()



import unittest

from flyers import dicts


class MyTestCase(unittest.TestCase):
    def test_something(self):
        m = dicts.Multimap()
        m['a'] = 1
        m['a'] = 2
        m['a'] = 3
        self.assertListEqual(m.get('a'), [1, 2, 3])

        del m['a']


if __name__ == '__main__':
    unittest.main()



import requests
from requests import Response

USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.163 ' \
             'Safari/537.36 '


class QueryIpWebsite(object):
    HTTPBIN = 'https://httpbin.org/ip'
    IPINFO = 'https://ipinfo.io/ip'


class ProxyType(object):
    SOCKS5 = 'SOCKS5'
    HTTP = 'HTTP'


def http_get(url, params=None, **kwargs) -> Response:
    if 'headers' in kwargs:
        kwargs['headers']['User-Agent'] = USER_AGENT
    else:
        kwargs['headers'] = {'User-Agent': USER_AGENT}
    return requests.get(url, params, **kwargs)


def http_post_file(url: str,
                   filename,
                   values: dict,
                   field_name: str = 'file') -> Response:
    with open(filename, 'rb') as f:
        file = {
            field_name: (filename, f.read())
        }
    return requests.post(url, files=file, data=values)


def build_proxies(proxy_type=ProxyType.HTTP, host='127.0.0.1', port=1080) -> dict:
    proxy_url = '{}:{}'.format(host, port)
    if proxy_type == ProxyType.HTTP:
        proxy = {
            "http": "http://{}".format(proxy_url),
            "https": "https://{}".format(proxy_url)
        }
    elif proxy_type == ProxyType.SOCKS5:
        proxy = {
            "http": "socks5://".format(proxy_url),
            "https": "socks5://{}".format(proxy_url)
        }
    else:
        raise TypeError('The value of type only has two types: socks5 and http, see ProxyType class')
    return proxy


def download(url: str, filename: str) -> None:
    b = requests.get(url).content
    with open(filename, 'wb+') as f:
        f.write(b)


def get_ip(ws: str = QueryIpWebsite.HTTPBIN):
    resp = http_get(ws)
    if ws == QueryIpWebsite.HTTPBIN:
        return resp.json()['origin']
    else:
        return resp.text



import unittest

from flyers import files


class MyTestCase(unittest.TestCase):
    def test_something(self):
        conf = files.read_ini_file("conf.ini")
        self.assertEqual(conf['server']['host'], '127.0.0.1')  # add assertion here
        self.assertEqual(conf['server']['port'], "8080")  # add assertion here


if __name__ == '__main__':
    unittest.main()



def is_true(s):
    if isinstance(s, bool):
        return s
    elif isinstance(s, int):
        return s == 1
    elif isinstance(s, str):
        return s == '1' or s == 'True' or s == 'true'
    else:
        return TypeError('The type is not expected: {}'.format(type(s)))



class StatusCodes(object):
    SUCCESS = 0


def success(data=None):
    return {'code': StatusCodes.SUCCESS, 'message': 'success', 'data': data}


def failure(code: int, message: str = None):
    return {'code': code, 'message': message, 'data': None}



import json


class DatabaseConf(object):
    def __init__(self,
                 database: str,
                 host: str = 'localhost',
                 port: int = 3306,
                 user: str = 'root',
                 password: str = ''):
        self.database = database
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def __str__(self):
        return str(self.__dict__)

    def json(self):
        return json.dumps(self.__dict__)

    @staticmethod
    def from_dict(d):
        return DatabaseConf(
            database=d['database'],
            host=d['host'],
            user=d['user'],
            port=d['port'],
            password=d['password']
        )


from . import pool
from . import mysql
from . import redisson



from flyers import strs


def check_null(s, name: str):
    if s is None:
        raise ValueError(f"The '{name}' cannot be null")


def check_blank(s: str, name: str):
    if strs.is_blank(s):
        raise ValueError(f"The '{name}' cannot be blank")


def check_list_empty(l: list):
    if len(l) == 0:
        raise ValueError(f"The list cannot be empty")


def check_dict_exists(d: dict, key: str):
    if key not in d:
        raise ValueError(f"The dict don't exists the '{key}' key")



import unittest
from flyers.logs import logger
from flyers import logs


class MyTestCase(unittest.TestCase):

    def test_something(self):
        logger.info("haha %s", "ha")
        self.assertFalse(logs.is_debug())
        logger.debug("debug message")

        logs.set_debug_level()
        self.assertTrue(logs.is_debug())

        if logs.is_debug():
            logger.debug("debug message")


if __name__ == '__main__':
    unittest.main()



def inspect_module():
    from flyers import database, spider
    from flyers.logs import get_rich_console_logger
    from rich import inspect
    from rich.console import Console
    from rich.text import Text
    from . import strs, files, dicts, lists, jsons, systems, texts, datetimes, concurrents, nets, assets, emails, logs
    import sys
    packages = {
        'assets': assets,
        'concurrents': concurrents,
        'datetimes': datetimes,
        'dicts': dicts,
        'emails': emails,
        'strs': strs,
        'files': files,
        'lists': lists,
        'jsons': jsons,
        'logs': logs,
        'systems': systems,
        'texts': texts,
        'nets': nets,
        'database.redisson': database.redisson,
        'spider.seleniums': spider.seleniums
    }

    console = Console()
    rich_logger = get_rich_console_logger('pkg')

    args = sys.argv
    if len(args) == 1:
        for k, v in packages.items():
            text = Text()
            text.append('{}   '.format(k), style='bold red')
            text.append(str(v), style='grey0')
            console.print(text)
    else:
        pn = args[1]
        if pn not in packages:
            rich_logger.error("The '{}' package is not present".format(pn))
            exit(-1)
        else:
            inspect(packages[pn], methods=True)


if __name__ == '__main__':
    inspect_module()



import json

from flyers import files


def to_json_str(d: dict, pretty: bool = False):
    if pretty:
        return json.dumps(d, indent=4)
    else:
        return json.dumps(d)


def to_json_file(d: dict, filename: str, pretty: bool = False):
    json_str = to_json_str(d, pretty)
    files.write_file(filename, json_str)


def from_file(filename: str):
    return from_json_str(files.read_file(filename))


def from_json_str(s: str, c=None):
    if not c:
        return json.loads(s)
    else:
        try:
            obj = c()
        except TypeError as e:
            raise Exception("Cannot initialize object '{}'".format(c.__name__), e)
        data = json.loads(s)
        for key in obj.__dict__.keys():
            setattr(obj, key, data[key])
        return obj


class JsonSerializable(object):

    def to_json_str(self):
        return json.dumps(self, default=lambda obj: obj.__dict__)

    @classmethod
    def from_json_str(cls, json_str):
        return from_json_str(json_str, cls)

    def __str__(self):
        return self.to_json_str()



from collections import OrderedDict

from flyers.collections import Pair


class Multimap(dict):
    def __setitem__(self, key, value):
        if key not in self:
            dict.__setitem__(self, key, [value])
        else:
            self[key].append(value)


class SortedOrderedDict(object):

    def __init__(self, other=()):
        self.od = OrderedDict(other)

    def __setitem__(self, field, value):
        self.od[field] = value

    def __getitem__(self, field):
        return self.od[field]

    def __delitem__(self, field):
        del self.od[field]

    def __contains__(self, field) -> bool:
        return field in self.od

    def keys(self) -> list:
        return list(self.od.keys())

    def values(self) -> list:
        return list(self.od.values())

    def items(self) -> list:
        return [Pair(each[0], each[1]) for each in self.od.items()]

    def sort_by_key(self):
        self.sort(comparator=lambda pair: pair.key)

    def sort_by_value(self):
        self.sort(comparator=lambda pair: pair.val)

    def sort(self, comparator):
        self.od = OrderedDict(sorted(self.od.items(), key=lambda x: comparator(Pair(x[0], x[1]))))

    def __str__(self):
        return self.od.__str__()

    def __repr__(self):
        return self.od.__repr__()


def get_or_none(d: dict, func):
    try:
        return func(d)
    except KeyError:
        return None


def get_or_default(d: dict, k: str, default_val):
    if k not in d:
        return default_val
    return d[k]



from functools import reduce
from random import Random

rand = Random()


def is_not_emtpy(l: list) -> bool:
    """
    Check whether the list is not empty
    :param l: The list
    :return:
    """
    return l is not None and len(l) != 0


def is_emtpy(l: list) -> bool:
    """
    Check whether the list is empty
    :param l: The list
    :return:
    """
    return not is_not_emtpy(l)


def get_random(l: list):
    """
    Gets random element from list
    :param l: The list
    :return: The random element in list
    """
    if len(l) == 0:
        return None
    return l[rand.randint(0, len(l) - 1)]


def for_each(l: list, func):
    for i in l:
        func(i)


def partition(l: list, num: int) -> list:
    return [l[i:i + num] for i in range(0, len(l), num)]


def flatten(l: list) -> list:
    return reduce(list.__add__, l)



import unittest
from flyers import strs


class MyTestCase(unittest.TestCase):

    def test_something(self):
        self.assertTrue(strs.is_blank('  '))
        self.assertFalse(strs.is_blank(' hello '))


if __name__ == '__main__':
    unittest.main()



import threading
import time

import schedule

from flyers import logger, datetimes, concurrents


class TimeUnit(object):
    SECONDS = 'seconds'
    MINUTES = 'minutes'
    HOURS = 'hours'
    DAYS = 'days'


# Beta
class BaseSchedulerTask(object):
    def __init__(self, name: str,
                 interval: int,
                 time_unit: str = TimeUnit.SECONDS,
                 run_at_start: bool = False, max_retries: int = 3,
                 loop_daly_seconds: int = 1):
        self.name = name
        self.interval = interval
        self.time_unit = time_unit
        self.run_at_start = run_at_start
        self.max_retries = max_retries
        self.loop_daly = loop_daly_seconds

        self.retries = 0
        self.start_time = datetimes.get_timestamp_ms()
        self.end_time = None
        self.end = False
        self.lock = threading.Lock()

    def start(self, *args, **kwargs):
        job = schedule.Job(self.interval, schedule.default_scheduler)
        job.unit = self.time_unit

        if self.run_at_start:
            self._execute(*args, **kwargs)
        job.do(self.__try_execute, *args, **kwargs)

        while not self.end:
            schedule.run_pending()
            time.sleep(self.loop_daly)

    def stop(self):
        with self.lock:
            self.end = True
            self.end_time = datetimes.get_timestamp_ms()

    def is_end(self):
        return self.end

    def get_cost_time(self) -> str:
        if self.end:
            et = self.end_time
        else:
            et = datetimes.get_timestamp_ms()
        return datetimes.format_duration(self.start_time, et)

    def __try_execute(self, *args, **kwargs):
        try:
            self._execute(*args, **kwargs)
        except Exception:
            self.retries += 1
            logger.exception('Task execute error, name = {}'.format(self.name))

            if self.retries > self.max_retries:
                raise Exception('Maximum number of retries exceeded')

    def _execute(self, *args, **kwargs):
        pass


def set_timeout(func, seconds: int):
    def runnable():
        time.sleep(seconds)
        func()

    concurrents.run_thread(runnable)



import pymysql

from flyers import logger


def transactional(conn: pymysql.Connection):
    def decorator(func):
        def wrapper(*args, **kwargs):
            with conn.cursor() as cursor:
                try:
                    ret = func(cursor, *args, **kwargs)
                    conn.commit()
                    return ret
                except:
                    logger.exception('execute transaction error: ')
                    conn.rollback()

        return wrapper

    return decorator



import diskcache as dc


# The super cache class
class Cache(object):
    def __init__(self, name):
        self.name = name

    def set(self, key: str, value):
        pass

    def get(self, key: str):
        pass


class DiskCache(Cache):
    def __init__(self, name):
        super().__init__(name)
        self.c = dc.Cache(self.name)

    def set(self, key: str, value):
        self.c[key] = value

    def get(self, key: str):
        return self.c.get(key)


class LoadingCache(DiskCache):

    def __init__(self, name):
        super().__init__(name)

    def get_or_load(self, key, load_func):
        """
        Gets the value from cache by cache and will load remote data automatically
        when the local cache is empty.
        :param key: The key of cache
        :param load_func: The function of loading data
        :return: The first param is cache value, the second value means whether to hit the cache.
        """
        if self.enabled_cache():
            result = self.get(key)
            if result:
                return result, True

        # loading data
        new_result = load_func()
        if self.enabled_cache():
            self.set(key, new_result)
        return new_result, False

    def enabled_cache(self):
        return True



import unittest
from flyers import systems
from flyers import logs


class MyTestCase(unittest.TestCase):

    def test_something(self):
        logs.set_debug_level()
        result = systems.run_shell_with_result("curl http://httpbin.org/get")
        if result == 0:
            logs.logger.info('success: {}'.format(result.stdout))
        else:
            logs.logger.info('failure: {}'.format(result.stderr))

        systems.run_shell('curl http://httpbin.org/get')
        systems.async_run_shell('curl http://httpbin.org/get', lambda ret: print(ret))


if __name__ == '__main__':
    unittest.main()



class Pair(object):
    def __init__(self, k, v):
        self.key = k
        self.val = v

    def __str__(self):
        return '{}->{}'.format(self.key, self.val)

    def __repr__(self):
        return self.__str__()



from redis import Redis

from flyers import lists


class RedisHashMap(object):

    def __init__(self, conn: Redis, key: str):
        self.conn = conn
        self.key = key

    def __setitem__(self, field, value):
        self.conn.hset(self.key, field, value)

    def __getitem__(self, field) -> str:
        return self.decode(self.conn.hget(self.key, field))

    def __delitem__(self, field):
        self.conn.hdel(self.key, field)

    def __contains__(self, field) -> bool:
        return self.conn.hexists(self.key, field)

    def delete(self):
        return self.conn.delete(self.key)

    def keys(self) -> list[str]:
        fields = self.conn.hkeys(self.key)
        if lists.is_emtpy(fields):
            return []
        return [self.decode(k) for k in fields]

    def values(self) -> list[str]:
        values = self.conn.hvals(self.key)
        if lists.is_emtpy(values):
            return []
        return [self.decode(k) for k in values]

    def items(self):
        field_values = self.conn.hgetall(self.key)
        result = {}
        for k, v in field_values.items():
            result[self.decode(k)] = self.decode(v)
        return result

    @staticmethod
    def decode(b: bytes):
        if not b:
            return None
        return bytes.decode(b)



from setuptools import setup, find_packages

setup(
    name='flyers',
    version='0.0.14',
    author='pushyzheng',
    author_email='1437876073@qq.com',
    url='https://github.com/pushyzheng',
    description=u'A collections of utils',
    packages=find_packages(),
    install_requires=['requests~=2.28.1', 'pymysql~=1.0.2', 'DBUtils~=3.0.2',
                      'selenium~=3.141.0', 'redis~=3.5.3', 'PyYAML~=5.4.1', 'rich~=10.16.2',
                      'schedule~=1.1.0', 'diskcache~=5.4.0']
)



from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Confirm, Prompt
from rich.table import Table

from flyers import jsons

DEFAULT_STYLE = 'cyan'

console = Console()


def print_table(title: str, columns: list, data_list: list):
    table = Table(title=title)

    for col in columns:
        if isinstance(col, str):
            table.add_column(col, style=DEFAULT_STYLE)
        elif isinstance(col, dict):
            table.add_column(col.get('content'), style=col.get('style') or DEFAULT_STYLE)
        else:
            raise Exception('The colum type is unsupported')

    for data in data_list:
        table.add_row(*data)

    console.print(table)


def print_to_json(d: dict):
    console.print_json(jsons.to_json_str(d))


def print_markdown(content: str):
    md = Markdown(content)
    console.print(md)


def print_panel(content: str, title: str = None, fit: bool = False):
    if not fit:
        console.print(Panel(content, title=title))
    else:
        console.print(Panel.fit(content, title=title))


def prompt_ask(prompt: str, default: str = None, choices: list[str] = None):
    return Prompt.ask(prompt, default=default, choices=choices)


def prompt_confirm(prompt: str):
    return Confirm.ask(prompt)



import csv
import os
import shutil
from configparser import ConfigParser

import yaml

from flyers import assets


class FileSizeUnit(object):
    B = 0
    KB = 1
    MB = 2
    GB = 3
    TB = 4
    PB = 5


class File(object):
    def __init__(self, path):
        assets.check_blank(path, 'path')
        self.path = path

    def is_file(self):
        return is_file(self.path)

    def is_dir(self):
        return not is_file(self.path)

    def rm(self):
        if self.is_file():
            rm_file(self.path)
        else:
            rmdir(self.path)

    def ls(self) -> (str, list):
        if not self.is_dir():
            return 'Not a directory: {}'.format(self.path), None
        return None, os.listdir(self.path)

    def read(self):
        return read_file(self.path)

    def write(self, new_content):
        return write_file(self.path, new_content)

    def append(self, new_lines):
        return append_file(self.path, new_lines)

    def md5(self):
        return get_md5(self.path)

    def size(self, unit):
        return get_file_size(self.path, unit)


def is_file(filename: str) -> bool:
    assets.check_blank(filename, 'filename')
    return os.path.isfile(filename)


def mv(src: str, dst: str):
    assets.check_blank(src, 'src')
    assets.check_blank(dst, 'dst')
    return os.rename(src, dst)


def read_file(filename: str, encoding='utf-8') -> str:
    assets.check_blank(filename, 'filename')
    with open(filename, encoding=encoding) as f:
        return f.read()


def read_lines(filename: str, encoding='utf-8') -> list[str]:
    assets.check_blank(filename, 'filename')
    with open(filename, encoding=encoding) as f:
        return f.readlines()


def read_and_splitlines(filename: str, encoding='utf-8') -> list[str]:
    assets.check_blank(filename, 'filename')
    with open(filename, encoding=encoding) as f:
        return f.read().splitlines()


def write_file(filename: str, data: str, encoding='utf-8'):
    assets.check_blank(filename, 'filename')
    with open(filename, 'w+', encoding=encoding) as f:
        f.write(data)


def append_file(filename: str, new_lines: str, encoding='utf-8'):
    assets.check_blank(filename, 'filename')
    with open(filename, 'a', encoding=encoding) as f:
        f.write(new_lines)


def update_file(filename: str, func):
    assets.check_blank(filename, 'filename')
    assets.check_blank(func, 'func')

    data = read_file(filename)
    new_data = func(data)
    write_file(filename, new_data)


def rm_file(filename: str):
    assets.check_blank(filename, 'filename')
    os.remove(filename)


def get_files(base_dir: str, endswith: str = None):
    assets.check_blank(base_dir, 'base_dir')

    result = []
    for f in os.listdir(base_dir):
        file_path = os.path.join(base_dir, f)
        if os.path.isfile(file_path) and (not endswith or f.endswith(endswith)):
            result.append(file_path)
    return result


def get_md5(filename: str) -> str:
    assets.check_blank(filename, 'filename')

    files_md5 = os.popen('md5 %s' % filename).read().strip()
    file_md5 = files_md5.replace('MD5 (%s) = ' % filename, '')
    return file_md5


def get_file_size(filename: str, unit: int) -> float:
    assets.check_blank(filename, 'filename')

    size_units = FileSizeUnit.__dict__
    if unit not in size_units.values():
        raise TypeError('No supports this unit')

    size = os.path.getsize(filename)
    for i in range(unit):
        size = size / float(1024)

    return size


def mkdir(d: str):
    os.mkdir(d)


def rmdir(d: str):
    shutil.rmtree(d)


def copy_dir(src: str, dst: str) -> None:
    for files in os.listdir(src):
        name = os.path.join(src, files)
        back_name = os.path.join(dst, files)
        if os.path.isfile(name):
            if os.path.isfile(back_name):
                if get_md5(name) != get_md5(back_name):
                    shutil.copy(name, back_name)
            else:
                shutil.copy(name, back_name)
        else:
            if not os.path.isdir(back_name):
                os.makedirs(back_name)
            copy_dir(name, back_name)


def read_yaml_file(filename: str) -> dict:
    assets.check_blank(filename, 'filename')
    return yaml.load(read_file(filename), yaml.FullLoader)


def read_csv_file(filename: str, delimiter: str = ' ', quote: str = '|'):
    assets.check_blank(filename, 'filename')
    with open(filename, newline='') as csvfile:
        csv.reader(csvfile, delimiter=delimiter, quotechar=quote)


def read_ini_file(filename: str) -> dict:
    assets.check_blank(filename, 'filename')
    conf = ConfigParser()
    conf.read(filename)
    result = {}
    for k, v in conf.items():
        item = {}
        for kk, vv in v.items():
            item.update({kk: vv})
        result.update({k: item})
    return result



import threading
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor

DEFAULT_MAX_WORKERS = 5


def asynchronous(func):
    def wrapper(*args):
        run_thread(target=func, args=args)

    return wrapper


def run_thread(target, args=(), blocking=False, timeout=10) -> None:
    t = threading.Thread(target=target, args=args)
    t.start()
    if blocking:
        t.join(timeout)


def execute_tasks_in_thread_pool(runnable,
                                 tasks: list,
                                 max_workers: int = DEFAULT_MAX_WORKERS) -> list:
    thread_pool = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix='flyers_default')
    futures = []
    for t in tasks:
        futures.append(thread_pool.submit(runnable, t))

    result = join_futures(futures)
    thread_pool.shutdown()
    return result


def execute_tasks_in_process_pool(runnable,
                                  tasks: list,
                                  max_workers: int = DEFAULT_MAX_WORKERS) -> list:
    process_pool = ProcessPoolExecutor(max_workers=max_workers)
    futures = []
    for t in tasks:
        futures.append(process_pool.submit(runnable, t))

    result = join_futures(futures)
    process_pool.shutdown()
    return result


def join_futures(futures: list):
    result = []
    for f in futures:
        result.append(f.result())
    return result



from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException


def scroll(driver, start: int, end: int):
    driver.execute_script('window.scrollTo({},{});'.format(start, end))


def screen_shot(driver, filename):
    driver.get_screenshot_as_file(filename)


def click_by_xpath(driver, xpath: str):
    click(driver, driver.find_element_by_xpath(xpath))


def click(driver, element):
    driver.execute_script("arguments[0].click();", element)


def wait_element(driver, xpath: str,
                 timeout_seconds: int = 10,
                 frequency: float = 0.5):
    return WebDriverWait(driver, timeout_seconds, frequency).until(
        EC.presence_of_element_located((By.XPATH, xpath)))


def has_element(driver, xpath: str):
    try:
        e = driver.find_element_by_xpath(xpath)
        return True, e
    except NoSuchElementException:
        return False, None
